﻿using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.MetaOptimization {
  /// <summary>
  /// TODO
  /// </summary>
  [Item("SolutionCacheAnalyzer", "")]
  [StorableClass]
  public sealed class SolutionCacheAnalyzer : SingleSuccessorOperator, IAnalyzer {

    public ValueLookupParameter<ResultCollection> ResultsParameter {
      get { return (ValueLookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    public ScopeTreeLookupParameter<ParameterConfigurationTree> ParameterConfigurationParameter {
      get { return (ScopeTreeLookupParameter<ParameterConfigurationTree>)Parameters["ParameterConfigurationTree"]; }
    }
    public ScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public LookupParameter<ConstrainedItemList<IProblem>> ProblemsParameter {
      get { return (LookupParameter<ConstrainedItemList<IProblem>>)Parameters[MetaOptimizationProblem.ProblemsParameterName]; }
    }
    public LookupParameter<IntValue> GenerationsParameter {
      get { return (LookupParameter<IntValue>)Parameters["Generations"]; }
    }
    public ValueParameter<BoolValue> StoreAllRunsParameter {
      get { return (ValueParameter<BoolValue>)Parameters["StoreAllRuns"]; }
    }

    public SolutionCacheAnalyzer()
      : base() {
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality", ""));
      Parameters.Add(new ValueLookupParameter<ResultCollection>("Results", ""));
      Parameters.Add(new ScopeTreeLookupParameter<ParameterConfigurationTree>("ParameterConfigurationTree", ""));
      Parameters.Add(new LookupParameter<ConstrainedItemList<IProblem>>(MetaOptimizationProblem.ProblemsParameterName, ""));
      Parameters.Add(new LookupParameter<IntValue>("Generations", ""));
      Parameters.Add(new ValueParameter<BoolValue>("StoreAllRuns", "If true all runs ever executed are stored. Otherwise only the runs from the latest generateion are stored for caching purposes.", new BoolValue(false)));
    }

    [StorableConstructor]
    private SolutionCacheAnalyzer(bool deserializing) : base(deserializing) { }
    private SolutionCacheAnalyzer(SolutionCacheAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new SolutionCacheAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      ResultCollection results = ResultsParameter.ActualValue;
      ItemArray<ParameterConfigurationTree> solutions = ParameterConfigurationParameter.ActualValue;
      ItemArray<DoubleValue> qualities = QualityParameter.ActualValue;
      ItemDictionary<StringValue, RunCollection> allRuns = ResultsParameter.ActualValue.ContainsKey("SolutionCache") ? (ItemDictionary<StringValue, RunCollection>)ResultsParameter.ActualValue["SolutionCache"].Value : new ItemDictionary<StringValue, RunCollection>();
      bool storeAllRuns = ((BoolValue)StoreAllRunsParameter.ActualValue).Value;

      if (!storeAllRuns) {
        allRuns.Clear();
      }

      foreach (var solution in solutions) {
        string key = solution.ParameterInfoString;
        bool first = false;
        if (allRuns.Count(x => x.Key.Value == key) == 0) {
          allRuns.Add(new StringValue(key), new RunCollection());
          first = true; // no other runs yet
        }

        var runCollection = allRuns.Single(x => x.Key.Value == key).Value;
        foreach (var run in solution.Runs) {
          if (!((BoolValue)run.Results["Meta.FromCache"]).Value || first) {
            run.Results["Meta.FromCache"] = new BoolValue(true);
            runCollection.Add(run);
          }
        }
      }

      if (!results.ContainsKey("SolutionCache")) {
        results.Add(new Result("SolutionCache", allRuns));
      }

      return base.Apply();
    }
  }
}
